@@ -1,13 +1,14 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2007 Dalius Dobravolskas <dalius@sandbox.lt>
+# Copyright (C) 2007-2013 Dalius Dobravolskas and Geoffrey T. Dairiki
 # All rights reserved.
 #
 # This software is licensed using the same licence as Trac:
 # http://trac.edgewall.org/wiki/TracLicense.
 #
-# Author: Dalius Dobravolskas <dalius@sandbox.lt>
-#
+# Original Author: Dalius Dobravolskas <dalius@sandbox.lt>
+# Current Maintainer: Jeff Dairiki <dairiki@dairiki.org>
+
 # Most probably you will want to add following lines to your configuration file:
 #
 #   [components]
@@ -22,10 +23,11 @@ import time
 import itertools
 
 from trac.core import *
-from trac.config import Option, BoolOption, IntOption
+from trac.config import Option, BoolOption, ListOption
 from trac.web.chrome import INavigationContributor, ITemplateProvider, add_stylesheet, add_script
 from trac.env import IEnvironmentSetupParticipant
 from trac.web.main import IRequestHandler, IAuthenticator
+from trac.perm import IPermissionGroupProvider, PermissionSystem
 from trac.web.session import DetachedSession
 try:
     from acct_mgr.web_ui import LoginModule
@@ -40,8 +42,11 @@ from openid.store.memstore import MemoryStore
 
 from openid.consumer import consumer
 from openid.extensions import sreg, pape, ax
-
 from openid import oidutil
+try:
+    from openid_teams.teams import TeamsRequest, TeamsResponse
+except ImportError:
+    TeamsRequest = TeamsResponse = None
 
 import socket
 import struct
@@ -73,7 +78,8 @@ class AuthOpenIdPlugin(Component):
         lastname='http://axschema.org/namePerson/last',
         )
 
-    implements(INavigationContributor, IRequestHandler, ITemplateProvider, IAuthenticator, IEnvironmentSetupParticipant)
+    implements(INavigationContributor, IRequestHandler, ITemplateProvider, IAuthenticator, IEnvironmentSetupParticipant,
+        IPermissionGroupProvider)
 
     # Do not declare options in the [trac] section.  We should not
     # be creating new declared options there, and we should not be
@@ -138,7 +144,7 @@ class AuthOpenIdPlugin(Component):
     trust_authname = BoolOption('openid', 'trust_authname', False,
             """WARNING: Only enable this if you know what this mean!
             This could make identity theft very easy if you do not control the OpenID provider!
-            Enabling this option makes the retrieved authname from the 
+            Enabling this option makes the retrieved authname from the
             OpenID provider authorative, i.e. it trusts the authname
             to be the unique username of the user. Enabling this disables
             the collission checking, so two different OpenID urls may
@@ -192,6 +198,12 @@ class AuthOpenIdPlugin(Component):
     custom_provider_size = Option('openid', 'custom_provider_size', 'small',
             """ Custom OpenId provider image size (small or large).""")
 
+    groups_to_request = ListOption('openid', 'groups_to_request', '',
+            doc=""" Which 'team names' to request via the OpenIDTeams extension.
+            To use this option you must have python-openid-teams installed.
+            """)
+
+
 
     def _get_masked_address(self, address):
         if self.check_ip:
@@ -243,6 +255,12 @@ class AuthOpenIdPlugin(Component):
         if type(store) is not MemoryStore:
             store.createTables()
 
+    # IPermissionGroupProvider methods
+    def get_permission_groups(self, username):
+        ds = DetachedSession(self.env, username)
+        return ds.get('openid.teams', '').split(',')
+
+
     # IEnvironmentSetupParticipant methods
 
     def environment_created(self):
@@ -339,7 +357,7 @@ class AuthOpenIdPlugin(Component):
             'custom_provider_size': self.custom_provider_size,
             }, None
 
-    def _get_session(self, req):
+    def _get_oidsession(self, req):
         """Returns a session dict that can store any kind of object."""
         try:
             return cPickle.loads(str(req.session[self.openid_session_key]))
@@ -358,11 +376,11 @@ class AuthOpenIdPlugin(Component):
 
         return base_url
 
-    def _commit_session(self, session, req):
-        req.session[self.openid_session_key] = str(cPickle.dumps(session))
+    def _commit_oidsession(self, oidsession, req):
+        req.session[self.openid_session_key] = str(cPickle.dumps(oidsession))
 
     def _get_consumer(self, req, db):
-        s = self._get_session(req)
+        s = self._get_oidsession(req)
         if 'id' not in s:
             s['id'] = req.session.sid
         store = self._getStore(db)
@@ -397,7 +415,7 @@ class AuthOpenIdPlugin(Component):
         immediate = 'immediate' in req.args
 
         db = self.env.get_db_cnx()
-        oidconsumer, session = self._get_consumer(req, db)
+        oidconsumer, oidsession = self._get_consumer(req, db)
         try:
             self.env.log.debug('beginning OpenID authentication.')
             request = oidconsumer.begin(openid_url)
@@ -437,7 +455,7 @@ class AuthOpenIdPlugin(Component):
                     'custom_provider_size': self.custom_provider_size,
                    }, None
             else:
-                self._commit_session(session, req)
+                self._commit_oidsession(oidsession, req)
                 # Then, ask the library to begin the authorization.
                 # Here we find out the identity server that will verify the
                 # user's identity, and get a token that allows us to
@@ -474,6 +492,16 @@ class AuthOpenIdPlugin(Component):
                     ax_request.add(attr_info)
                 request.addExtension(ax_request)
 
+                if self.groups_to_request:
+                    if not TeamsRequest:
+                        self.env.log.error(
+                            'The python-openid-teams package is not installed.'
+                            ' The groups_to_request configuration option will'
+                            ' be ignored.')
+                    else:
+                        request.addExtension(
+                            TeamsRequest(requested=self.groups_to_request))
+
                 trust_root = self._get_trust_root(req)
                 if self.absolute_trust_root:
                     trust_root += '/'
@@ -500,7 +528,7 @@ class AuthOpenIdPlugin(Component):
         """Handle the redirect from the OpenID server.
         """
         db = self.env.get_db_cnx()
-        oidconsumer, session = self._get_consumer(req, db)
+        oidconsumer, oidsession = self._get_consumer(req, db)
 
         # Ask the library to check the response that the server sent
         # us.  Status is a code indicating the response type. info is
@@ -522,6 +550,8 @@ class AuthOpenIdPlugin(Component):
             # the verification.
             css_class = 'alert'
 
+            session_attr = {}           # attributes for new "user"
+
             # This is a successful verification attempt. If this
             # was a real application, we would do our login,
             # comment posting, etc. here.
@@ -550,6 +580,16 @@ class AuthOpenIdPlugin(Component):
 
             nickname = sreg_info.get('nickname')
 
+            if self.groups_to_request and TeamsResponse:
+                teams_response = TeamsResponse.fromSuccessResponse(info)
+                if teams_response:
+                    # be careful not to make user a member of any trac groups
+                    # not named in groups_to_request
+                    teams = set(teams_response.teams
+                                ).intersection(self.groups_to_request)
+                    if teams:
+                        session_attr['openid.teams'] = ','.join(teams)
+
             if self.strip_protocol:
                 remote_user = remote_user[remote_user.find('://')+3:]
             if self.strip_trailing_slash and remote_user[-1] == '/':
@@ -577,13 +617,14 @@ class AuthOpenIdPlugin(Component):
                     if item.match(remote_user):
                         allowed = False
                         self.env.log.debug("User black-listed.")
-            if allowed and email and self.re_email_white_list:
-                self.env.log.debug("Filtering email '%s' through email white-list." % email)
+            if allowed and self.re_email_white_list:
+                self.env.log.debug("Filtering email %r through email white-list." % email)
                 allowed = False
-                for item in self.re_email_white_list:
-                    if not allowed and item.match(email):
-                        allowed = True
-                        self.env.log.debug("User email white-listed.")
+                if email:
+                    for item in self.re_email_white_list:
+                        if not allowed and item.match(email):
+                            allowed = True
+                            self.env.log.debug("User email white-listed.")
 
             if allowed and self.check_list:
                 allowed = False
@@ -613,20 +654,20 @@ class AuthOpenIdPlugin(Component):
                 if cookie_lifetime > 0:
                     req.outcookie['trac_auth']['expires'] = cookie_lifetime
 
-                req.session[self.openid_session_identity_url_key] = info.identity_url
+                session_attr[self.openid_session_identity_url_key] = info.identity_url
                 if email:
-                    req.session['email'] = email
+                    session_attr['email'] = email
                 if fullname:
-                    req.session['name'] = fullname
+                    session_attr['name'] = fullname
 
-                self._commit_session(session, req)
+                self._commit_oidsession(oidsession, req)
 
                 if self.check_list and self.check_list_username:
                     authname = cl_username
                 elif self.use_nickname_as_authname and nickname:
                     authname = nickname
-                elif req.session.get('name'):
-                    authname = req.session['name']
+                elif session_attr.get('name'):
+                    authname = session_attr['name']
                     if self.combined_username:
                         authname = '%s <%s>' % (authname, remote_user)
                 else:
@@ -636,11 +677,13 @@ class AuthOpenIdPlugin(Component):
                 if self.lowercase_authname:
                     authname = authname.lower()
 
-                if not self.trust_authname:
+                if self.trust_authname:
+                    ds = DetachedSession(self.env, authname)
+                else:
                     # Make authname unique in case of collisions
                     #
                     # XXX: We ought to first look for an existing authenticated
-                    # ssession with matching identity_url, and just use that
+                    # session with matching identity_url, and just use that
                     # for the authid.  (E.g. what if the user changes his
                     # fullname at the openid provider?)  However, trac does
                     # not seem to provide an API for searching sessions other
@@ -651,16 +694,46 @@ class AuthOpenIdPlugin(Component):
                         for attempt in itertools.count(2):
                             yield "%s (%d)" % (base, attempt)
 
+                    existing_users_and_groups = set(
+                        user
+                        for user, perm
+                        in PermissionSystem(self.env).get_all_permissions())
+
                     for authname in authnames(authname):
                         ds = DetachedSession(self.env, authname)
                         if ds.last_visit == 0 and len(ds) == 0:
                             # At least in 0.12.2, this mean no session exists.
+                            if authname in existing_users_and_groups:
+                                # Permissions are already defined for this user
+                                continue
                             break
                         ds_identity = ds.get(self.openid_session_identity_url_key)
                         if ds_identity == info.identity_url:
                             # No collision
                             break
 
+                if ds and (ds.last_visit != 0 or len(ds) > 0):
+                    # The user already exists, update team membership
+                    # XXX: Should also update name and/or email? (This would
+                    # be an API change.)
+                    for name in ['openid.teams']:
+                        if name in session_attr:
+                            ds[name] = session_attr[name]
+                        elif name in ds:
+                            del ds[name]
+                    ds.save()
+                else:
+                    # We are creating a new "user".  Set attributes on the
+                    # current anonymous session.  It will be promoted to
+                    # the new authenticated session on the next request
+                    # (by Session.__init__).
+                    #
+                    # NB: avoid dict.update here to ensure that
+                    # DetachedSession.__getitem__ gets a chance to
+                    # normalize values
+                    for name, value in session_attr.items():
+                        req.session[name] = value
+
                 req.authname = authname
 
                 db = self.env.get_db_cnx()
@@ -691,7 +764,7 @@ class AuthOpenIdPlugin(Component):
             # information in a log.
             message = 'Verification failed.'
 
-        self._commit_session(session, req)
+        self._commit_oidsession(oidsession, req)
 
         add_stylesheet(req, 'authopenid/css/openid.css')
         add_script(req, 'authopenid/js/openid-jquery.js')
